package AITB_2022.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.complex.HumanDetector;
import adf.core.component.communication.CommunicationMessage;
import adf.core.debug.DefaultLogger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import org.apache.log4j.Logger;
import rescuecore2.standard.entities.*;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.standard.entities.Civilian;




public class AITATHumanDetector extends HumanDetector {

  private Clustering clustering;

  private EntityID result;

  private Logger logger;

  private MessageManager messageManager;

  private double lastX = -1.0;//1step前のエージェント自身のX座標

  private double lastY = -1.0;//1step前のエージェント自身のY座標

  private int police_coll_flag = 0;//土木隊を呼んだときのクールタイム

  //通信用のアクション番号
  final private int HELP_BLOCKADE = 6;//瓦礫で動けない状態

//private Collection<EntityID> checklist;

  public AITATHumanDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    this.clustering = moduleManager.getModule("AITATHumanDetector.Clustering",
        "adf.impl.module.algorithm.KMeansClustering");
    registerModule(this.clustering);
    //this.checklist = new ArrayList<EntityID>();
  }


  @Override
  public HumanDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    super.updateInfo(messageManager);
    if (this.getCountUpdateInfo() > 1) {
      return this;
    }
    this.messageManager = messageManager; // messageManagerの定義が必要
    // そのステップで受信した(前ステップで送信された)メッセージ一覧
    List<CommunicationMessage> messages = messageManager.getReceivedMessageList();
    this.logger.debug("Time:"+this.agentInfo.getTime()+" mess: "+messages);
    return this;
  }


  @Override
  public HumanDetector calc() {
    if(this.agentInfo.getTime() > 3){ //最初の3stepは救急隊が止まってるから3stepまではおこなわない
      //自分が瓦礫で止まっていたら土木隊に送信
      this.message_police((AmbulanceTeam)this.agentInfo.me(),this.agentInfo.getPosition());
    }

    // 市民を運んでいるかどうか
    Human transportHuman = this.agentInfo.someoneOnBoard(); // 運んでいる市民のエンティティを取得
    if (transportHuman != null) { // 前のステップで運んでいたら
      logger.debug("someoneOnBoard:" + transportHuman);
      this.result = transportHuman.getID(); // resultに取得した市民のIDを代入
      return this;
    }
    if (this.result != null) { // 前のステップで運んでいたら
      Human target = (Human) this.worldInfo.getEntity(this.result); // 市民のエンティティを取得
      if (!isValidHuman(target)) { // 運ぶ必要がないとき
        logger.debug("Invalid Human:" + target + " ==>reset target");
        this.result = null; //
      }
    }
      // 市民を運んでいなかったらcalcTargetで市民を探す
    if (this.result == null) {
      this.result = calcTarget();
    }

    //知覚している市民のEntityIDコレクションを取得
    Collection<EntityID> ids = this.worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN);

    // for文で埋没されている市民を見つけたらWorldInfoに情報を追加
    for (EntityID id : ids)//拡張forで市民のEntityIDをまわす
    {
      StandardEntity entity = this.worldInfo.getEntity(id);//idからentityを取得

      Civilian civ = (Civilian) entity;//ダウンキャスト

      if (civ != null) { // civにデータが入っているとき
        // 死んでいたらスルー
        if (!civ.isHPDefined() && civ.getHP() == 0) continue;
        // 市民が埋没していたら
        if (civ.isBuriednessDefined() && civ.getBuriedness() > 0) {
          this.logger.debug("Time:" + this.agentInfo.getTime() + " 埋没度の取得 "+civ.getBuriedness()+" civ " + civ);
          CommunicationMessage mm = new MessageCivilian(false,civ);
          this.messageManager.addMessage(mm,false);
          this.logger.debug("Time:" + this.agentInfo.getTime() + " 消防隊を呼ぶ");
          System.out.println("消防隊を呼ぶ");
          //this.checklist.add(civ.getID()); // 一度呼んだ市民のIDを追加
        }
      }
    }
    return this;
  }

  // 現在、瓦礫で身動きが取れないかを判断し、真なら土木隊に送信する処理
  //第一引数は自身のEntity(AmbulanceTeam), 第二引数は現在場所のID
  public void message_police(AmbulanceTeam atMeEntity, EntityID nowPositionID){
    //エージェント自身が現在居る場所のエンティティをIDから取得
    StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);

    double nowX = this.agentInfo.getX();//現在のX座標を取得
    double nowY = this.agentInfo.getY();//現在のY座標を取得
    this.logger.debug("Time:"+this.agentInfo.getTime()+" nowPositionEntity "+nowPositionEntity);

    double compDistanceX = Math.abs((nowX - this.lastX));//現在のX座標と1step前のX座標との差
    double compDistanceY = Math.abs((nowY - this.lastY));//現在のY座標と1step前のY座標との差

    this.logger.debug("nowX "+nowX+" lastX "+lastX+" compDistanceX "+compDistanceX);
    this.logger.debug("nowY "+nowY+" lastY "+lastY+" compDistanceY "+compDistanceY);

    //もしエージェントが現在いる場所が道路&&1step前と場所が全く変わっていない&&土木隊を呼ぶクールタイムが0のとき
    if((nowPositionEntity instanceof Road) && ((compDistanceX < 3.0)&&(compDistanceY < 3.0)) && (this.police_coll_flag == 0)){
      this.police_coll_flag = 10;//土木隊を呼ぶクールタイムを設定
      this.logger.debug("現在、Road居てさっきとエージェントの場所が変わってなくてクールタイム中でもない。つまり瓦礫で止まってます");

      //救急隊が埋まったという情報を送信
      messageManager.addMessage(new MessageAmbulanceTeam(false,atMeEntity,HELP_BLOCKADE,nowPositionID),false);
      System.out.println("救急隊が土木隊を呼ぶ");

    }else if(this.police_coll_flag != 0){//もし土木隊を呼ぶクールタイムが0でないなら
      this.police_coll_flag -= 1;//土木隊を呼ぶクールタイムを減少させる
    }
    this.lastX = nowX;//今のXをlastXに代入
    this.lastY = nowY;//今のYをlastYに代入
  }

  private EntityID calcTarget() {
    List<Human> rescueTargets = filterRescueTargets(
        this.worldInfo.getEntitiesOfType(CIVILIAN));
    List<Human> rescueTargetsInCluster = filterInCluster(rescueTargets);
    List<Human> targets = rescueTargetsInCluster;
    if (targets.isEmpty())
      targets = rescueTargets;

    logger.debug("Targets:" + targets);
    if (!targets.isEmpty()) {
      targets.sort(new DistanceSorter(this.worldInfo, this.agentInfo.me()));
      Human selected = targets.get(0);
      logger.debug("Selected:" + selected);
      return selected.getID();
    }
    return null;
  }


  @Override
  public EntityID getTarget() {
    return this.result;
  }


  private List<Human>
      filterRescueTargets(Collection<? extends StandardEntity> list) {
    List<Human> rescueTargets = new ArrayList<>();
    for (StandardEntity next : list) {
      if (!(next instanceof Human))
        continue;
      Human h = (Human) next;
      if (!isValidHuman(h))
        continue;
      if (h.getBuriedness() > 0)
        continue;
      rescueTargets.add(h);
    }
    return rescueTargets;
  }


  private List<Human>
      filterInCluster(Collection<? extends StandardEntity> entities) {
    int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    List<Human> filter = new ArrayList<>();
    HashSet<StandardEntity> inCluster = new HashSet<>(
        clustering.getClusterEntities(clusterIndex));
    for (StandardEntity next : entities) {
      if (!(next instanceof Human))
        continue;
      Human h = (Human) next;
      if (!h.isPositionDefined())
        continue;
      StandardEntity position = this.worldInfo.getPosition(h);
      if (position == null)
        continue;
      if (!inCluster.contains(position))
        continue;
      filter.add(h);
    }
    return filter;

  }

  private class DistanceSorter implements Comparator<StandardEntity> {

    private StandardEntity reference;
    private WorldInfo worldInfo;

    DistanceSorter(WorldInfo wi, StandardEntity reference) {
      this.reference = reference;
      this.worldInfo = wi;
    }


    public int compare(StandardEntity a, StandardEntity b) {
      int d1 = this.worldInfo.getDistance(this.reference, a);
      int d2 = this.worldInfo.getDistance(this.reference, b);
      return d1 - d2;
    }
  }

  // 運ぶ必要があるか判断し，そこが建物か道路かを判断する
  private boolean isValidHuman(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    if (!target.isHPDefined() || target.getHP() == 0) // HPが定義されていない　もしくは　HPが0　のとき
      return false;
    if (!target.isPositionDefined()) // 市民がどこにいるかわからないとき
      return false;
    if (!target.isDamageDefined() || target.getDamage() == 0) // 市民がダメージを受けていないとき
      return false;
    if (!target.isBuriednessDefined()) // 埋没度が定義されていないとき
      return false;

    StandardEntity position = worldInfo.getPosition(target); // 市民がいる建物もしくは道路の取得
    if (position == null) // 市民が建物内もしくは道路にいないとき
      return false;

    StandardEntityURN positionURN = position.getStandardURN(); //  建物もしくは道路の識別子を取得
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM) // 取得した識別子が休憩所　もしくは　救急隊司令塔　のとき
      return false;

    return true;
  }
}