package AITB_2022.module.complex;

import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.communication.standard.bundle.information.MessageRoad;
import adf.core.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.core.agent.info.*;
import adf.core.component.module.complex.HumanDetector;
import adf.core.component.communication.CommunicationMessage;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import adf.core.component.module.algorithm.Clustering;
import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;

import static adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam.ACTION_LOAD;
import static java.lang.Double.POSITIVE_INFINITY;
import static java.lang.Integer.MIN_VALUE;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;

//timeがステップ数 targetが市民のエンティティID 
// 行動(掘り起こし)対象(市民)を決定するクラス
public class AITFBHumanDetector extends HumanDetector
{
    private EntityID result;
    private Clustering clustering;
    private MessageManager messageManager;
    private Logger logger;

    // 1step前の情報を保持するフィールド群
    private EntityID lastID = null;
    private double lastX = -1.0;//1step前のエージェント自身のX座標
    private double lastY = -1.0;//1step前のエージェント自身のY座標

    private int police_coll_flag = 0;//土木隊を呼んだときのクールタイム

    //通信用のアクション番号
    final private int HELP_BLOCKADE = 6;//瓦礫で動けない状態

    //エージェントごとの掘り起こし優先度
    private int civ_priority = 1;
    private int at_priority = 2;
    private int fb_priority = 2;
    private int pf_priority = 2;

    public AITFBHumanDetector(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);
        this.clustering = mm.getModule("SampleHumanDetector.Clustering", "adf.impl.module.algorithm.KMeansClustering");
        this.registerModule(this.clustering);

        this.logger = DefaultLogger.getLogger(agentInfo.me());
    }

    //行動対象をreturnするメソッド
    @Override
    public EntityID getTarget()
    {
        return this.result;
    }
    
    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if (this.getCountUpdateInfo() > 1) {
            return this;
        }
        this.messageManager = messageManager;
        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages = messageManager.getReceivedMessageList();
        this.logger.debug("Time:"+this.agentInfo.getTime()+" mess: "+messages);

        return this;
    }

    //行動対象を計算するメソッド
    @Override
    public HumanDetector calc()
    {
        //このエージェントが担当するクラスタの番号を取得し，idxに代入
        int idx = this.clustering.getClusterIndex(this.agentInfo.getID());
        //idx番のクラスタに属するEntityをすべて取得し，myClusterEntitiesに代入
        Collection<StandardEntity> myClusterEntities = this.clustering.getClusterEntities(idx);
        //シミュレーションの1ステップ目のみ、このエージェントが担当するクラスタに属するEntityを表示
        if (this.agentInfo.getTime() == 1) logger.debug(myClusterEntities);
        
        EntityID me = this.agentInfo.getID();//エージェント自身のIDを取得
        //知覚している市民,消防隊,救急隊,土木隊のEntityIDコレクションを取得
        Collection<EntityID> ids = this.worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN,AMBULANCE_TEAM,FIRE_BRIGADE,POLICE_FORCE);

        //Collection<StandardEntity> entitys = this.worldInfo.getEntitiesOfType(StandardEntityURN.CIVILIAN,AMBULANCE_TEAM,FIRE_BRIGADE,POLICE_FORCE);//知覚しているHumanのEntityIDコレクションを取得
        //this.logger.debug("time"+this.agentInfo.getTime()+" Civilian_entities "+entitys);
        //logger.debug("time "+this.agentInfo.getTime()+" change_id "+this.worldInfo.getChanged().getChangedEntities());
        
        //最初の3stepは消防隊が止まってる？から3stepまでは行わない
        if(this.agentInfo.getTime() > 3){
            //自分が瓦礫で止まっているかを判定して真なら土木隊に送信
            this.message_police((FireBrigade)this.agentInfo.me(),this.agentInfo.getPosition());
        }

        EntityID nearestID = null;//一番近いHumanのIDを格納する変数 初期値はnull
        double nearestDistance = Double.POSITIVE_INFINITY;//nearesDistanceに初期値としてDoubleで最も大きい値を代入
        int high_digup_priority = MIN_VALUE;//探索中の現在のresultの掘り起こし優先度を格納する変数
        int digup_priority = 0;//Humanの掘り起こし優先度を格納する変数

        //stepが60より大きい && 現在掘り起こしていない ならば
        if(this.agentInfo.getTime()>60 && !isDigup()){
            change_digup_priority();// 掘り起こし優先度の設定を変更
        }

        //救助対象を探索
        for (EntityID id : ids)//拡張forでHumanのEntityIDをまわす
        {
            StandardEntity entity = this.worldInfo.getEntity(id);//idからentityを取得

            //entityに紐ずく掘り起こし優先度を取得
            switch(entity.getStandardURN()){
                case CIVILIAN://entityが市民であれば
                    digup_priority = this.civ_priority;
                    break;
                case AMBULANCE_TEAM://entityが救助隊であれば
                    digup_priority = this.at_priority;
                    break;
                case FIRE_BRIGADE://entityが消防隊であれば
                    digup_priority = this.fb_priority;
                    break;
                case POLICE_FORCE://entityが土木隊であれば
                    digup_priority = this.pf_priority;
                    break;
            }

            //もし掘り起こし優先度が0なら掘り起こさずスルー
            if(digup_priority==0) continue;

            Human hum = (Human) entity;//ダウンキャスト
            
            //クラスタ内のEntityに市民の場所のEntityが含まれていなかったらスルー
            if(!myClusterEntities.contains(this.worldInfo.getEntity(hum.getPosition()))) continue;

            boolean isAlive = hum.isHPDefined() && hum.getHP() > 0;//生きてたらtrue
            //生きていないならスルー
            if (!isAlive) continue;

            boolean isBury = hum.isBuriednessDefined() && hum.getBuriedness() > 0;//埋まってたらtrue
            //埋まっていないならスルー
            if(!isBury) continue;

            //logger.debug("time"+this.agentInfo.getTime()+" "+hum+ " digup_priority "+digup_priority+" 埋没度 "+hum.getBuriedness());

            //掘り起こし優先度が同一なら
            if(high_digup_priority == digup_priority){
                double distance = this.worldInfo.getDistance(me, id);//humanとの間の距離を取得
                if (distance < nearestDistance)//もし現在保持している最小距離より距離が小さいなら
                {
                    nearestID = id;//IDを代入
                    nearestDistance = distance;//距離を一番近い距離として代入
                }
            //掘り起こし優先度が上位なら
            }else if(high_digup_priority < digup_priority){
                high_digup_priority = digup_priority;//掘り起こし優先度を代入
                nearestID = id;//IDを代入
                nearestDistance = this.worldInfo.getDistance(me, id);//距離を一番近い距離として代入
            }
        }
        this.result = nearestID;//一番距離が近いIDを救助対象として代入
        //if(this.result != null) logger.debug("time"+this.agentInfo.getTime()+" 救助対象は "+this.worldInfo.getEntity(this.result)+" に決まった");
        

        //1step前の救助対象のIDからentityを取得
        StandardEntity lastEntity = this.worldInfo.getEntity(lastID);
        
        //lastEntityを掘り起こし終わった && lastEntityインスタンスがCivilianクラス ならば
        if(isDigupfinish(this.result) && lastEntity instanceof Civilian){
        
            Civilian civ = (Civilian) lastEntity;//ダウンキャスト
            this.logger.debug("救急隊は市民" +lastID+ "を連れてけ" +" resultは"+this.result);
            messageManager.addMessage(new MessageCivilian(false,civ),false);//救助した市民情報を送信

        }

        this.lastID = this.result;//最後に現在の救助対象IDをlastIDとして格納
        this.lastX = this.agentInfo.getX();//今のXをlastXに格納
        this.lastY = this.agentInfo.getY();//今のYをlastYに格納

        return this;
    }
    
    // 現在、瓦礫で身動きが取れないかを判断し、真なら土木隊に送信する処理
    private void message_police(FireBrigade fbMeEntity, EntityID nowPositionID){//第一引数は自身のEntity(FireBrigade), 第二引数は現在場所のID
        //もしエージェントが瓦礫で止まっていて、土木隊を呼ぶクールタイムが0のとき
        if(isStop_broken() && (this.police_coll_flag == 0)){
            this.police_coll_flag = 3;//土木隊を呼ぶクールタイムを設定
            this.logger.debug("現在、Road居てさっきとエージェントの場所が変わってなくてクールタイム中でもない。瓦礫で止まってます");
            System.out.println("time: "+this.agentInfo.getTime()+" 瓦礫で止まってます");

            //消防隊が埋まったという情報を送信
            messageManager.addMessage(new MessageFireBrigade(false,fbMeEntity,HELP_BLOCKADE,nowPositionID),false);

        }else if(this.police_coll_flag != 0){//もし土木隊を呼ぶクールタイムが0でないなら
            this.police_coll_flag -= 1;//土木隊を呼ぶクールタイムを減少させる
        }
    }

    // 掘り起こしているかを判定するメソッド
    private boolean isDigup(){
        EntityID nowPositionID = this.agentInfo.getPosition();//エージェント自身の現在場所のID
        StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);//エージェント自身が現在居る場所のエンティティをIDから取得
        double[] pos_dif = compDistanceXY();//エージェント自身の現在の位置と1step前の位置の差を取得{x,y}

        if(this.lastID==null) return false;//1step前の値がnull->掘り起こしていない

        Human lastEntity = (Human)this.worldInfo.getEntity(this.lastID);//1step前の救助対象を取得
        //もしエージェントが現在いる場所が建物で、1step前と場所が全く変わっておらず、lastEntityの埋没度が0より大きいとき
        if((nowPositionEntity instanceof Building) && ((pos_dif[0] < 3.0)&&(pos_dif[1] < 3.0)) && lastEntity.getBuriedness() > 0){
            return true;
        }else{
            return false;
        }
    }
    // 掘り起こし終わったかを判定するメソッド
    private boolean isDigupfinish(EntityID nowRescueTargetID){
        Human lastEntity = (Human)this.worldInfo.getEntity(this.lastID);//1step前の救助対象を取得
        // 1step前の救助対象IDがnullでない && 1step前の救助対象IDが今の救助対象IDと異なる && 1step前の救助対象の埋没度が0
        if(lastID != null && lastID != nowRescueTargetID && lastEntity.getBuriedness() == 0){
            return true;
        }else{
            return false;
        }
    }

    // エージェントが瓦礫で止まっているかを判定するメソッド
    private boolean isStop_broken(){
        EntityID nowPositionID = this.agentInfo.getPosition();//エージェント自身が現在居る場所のID
        StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);//エージェント自身が現在居るエリアのIDからエリアのエンティティを取得
        double[] pos_dif = compDistanceXY();//エージェント自身の現在の位置と1step前の位置の差を取得{x,y}

        //もしエージェントが現在いる場所が道路で、1step前と場所が全く変わっていないとき
        if((nowPositionEntity instanceof Road) && ((pos_dif[0] < 3.0)&&(pos_dif[1] < 3.0))){
            return true;
        }else{
            return false;
        }
    }

    // エージェント自身の現在の位置と1step前の位置の差を出力するメソッド
    private double[] compDistanceXY(){
        double nowX = this.agentInfo.getX();//現在のX座標を取得
        double nowY = this.agentInfo.getY();//現在のY座標を取得
        double posX_dif = Math.abs((nowX - this.lastX));//現在のX座標と1step前のX座標との差
        double posY_dif = Math.abs((nowY - this.lastY));//現在のY座標と1step前のY座標との差
        double[] pos_dif = {posX_dif,posY_dif};//X座標の差とY座標の差を持つ配列
        return pos_dif;
    }

    // 掘り起こし優先度の設定を変更するメソッド
    private void change_digup_priority(){
        this.civ_priority = 2;
        this.at_priority = 1;
        this.fb_priority = 1;
        this.pf_priority = 1;
        //logger.debug("掘り起こし優先度を変更しました");
    }
}