package AITB_2022.module.complex;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.StandardMessagePriority;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.info.*;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.complex.Search;
import adf.core.component.module.algorithm.Clustering;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import java.util.stream.Collectors;

public class AITPFSearch extends Search {
    private EntityID result;
    private Random random = new Random(1);
    private Clustering clustering;
    private List<EntityID> clonelist = new ArrayList<>();
    private List<EntityID> list = new ArrayList<>();
    private MessageManager messageManager;
    private Set<EntityID> rmSet = new HashSet<>();
    private Set<EntityID> searchedSet = new HashSet<>();

    public AITPFSearch(
        AgentInfo ai, WorldInfo wi, ScenarioInfo si,
        ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);

        String clusteringKey = "SampleSearch.Clustering.Police";
        this.clustering =
            mm.getModule(clusteringKey,
                    "adf.sample.module.algorithm.SampleKMeans");
        this.registerModule(this.clustering);
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    private void rmbuilding() {
        EntityID me = this.agentInfo.getID();
        int idx = this.clustering.getClusterIndex(me);
        Collection<EntityID> cluster = this.clustering.getClusterEntityIDs(idx);
        list = new ArrayList<>(cluster);
        if (list.size() > 0){
            // 建物のリストを作成
            for (EntityID build : list) {
                StandardEntity entity = this.worldInfo.getEntity(build);
                if(entity instanceof Building) rmSet.add(build);
            }
            // 建物を削除
            for (EntityID l : rmSet) {
                list.remove(l);
            }
            // リストのコピーを作成
            clonelist = new ArrayList<>(list);
        }
    }

    @Override
    public Search updateInfo(MessageManager mm)
    {
        super.updateInfo(mm);
        if (this.getCountUpdateInfo() > 1) return this;

        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages =
                mm.getReceivedMessageList();
        this.messageManager = mm;
        return this;

    }

    @Override
    public Search calc()
    {
        EntityID me = this.agentInfo.getID();
        int idx = this.clustering.getClusterIndex(me);
        if (idx == -1) {
        return this;
        }

        Collection<EntityID> cluster =
                this.clustering.getClusterEntityIDs(idx);
        if (cluster == null) {
        return this;
        }

        // エージェントが現在位置する建物や道路のIDを取得する
        EntityID position = this.agentInfo.getPosition();
        
        // 建物を除去するためのセットを設定
        if(rmSet.size() < 1) rmbuilding();

        // 対象がない時、対象を決定
        if(result == null) {
            if(clonelist.size() > 0){ 
                list = new ArrayList<>(clonelist);
            }else{
                list = new ArrayList<>(cluster);
            }
            // 擬似乱数を用いてランダムに選択します
            int n = list.size();
            int r = random.nextInt(n);
            this.result = list.get(r);
        }

        // 対象に辿り着くまで，対象を変更しない
        if (!this.getTarget().equals(position)) {
        return this;
        }

        // 探索済みの (辿り着いたことがある) 対象を候補から除外する
        if (this.getTarget().equals(position)) {
        this.searchedSet.add(this.getTarget());
        list.removeAll(this.searchedSet);
        // すべての対象が探索済みの場合，候補をリセットする
        if (list.size() <= 0) {
            this.searchedSet.clear();
            if(clonelist.size() > 0){ 
                list = new ArrayList<>(clonelist);
            }else{
                list = new ArrayList<>(cluster);
            }
        }
        // 擬似乱数を用いてランダムに選択します
        int n = this.list.size();
        int r = this.random.nextInt(n);
        this.result = list.get(r);
        return this;
        }
        
        return this;
    }
}