package AITB_2022.module.complex;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.StandardMessagePriority;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.info.*;
import adf.core.agent.precompute.PrecomputeData;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.complex.Search;
import adf.core.component.module.algorithm.Clustering;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import java.util.stream.Collectors;

import static java.lang.Double.POSITIVE_INFINITY;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;

public class AITSearch extends Search
{
  private EntityID result;
  private Random random = new Random(1);
  private Clustering clustering;
  private Set<EntityID> searchedSet = new HashSet<>();
  private List<EntityID> list;
  private MessageManager messageManager;
  private static final int ACTION_LOAD = CommandAmbulance.ACTION_LOAD;

  public AITSearch(
          AgentInfo ai, WorldInfo wi, ScenarioInfo si,
          ModuleManager mm, DevelopData dd)
  {
    super(ai, wi, si, mm, dd);

    String clusteringKey = "SampleSearch.Clustering";
    switch (ai.me().getStandardURN())
    {
      case FIRE_BRIGADE:
        clusteringKey += ".Fire";
        break;
      case AMBULANCE_TEAM:
        clusteringKey += ".Ambulance";
        break;
      case POLICE_FORCE:
        clusteringKey += ".Police";
        break;
    }

    this.clustering =
            mm.getModule(clusteringKey,
                    "adf.sample.module.algorithm.SampleKMeans");
    this.registerModule(this.clustering);
  }

  @Override
  public EntityID getTarget()
  {
    return this.result;
  }

  @Override
  public Search updateInfo(MessageManager mm)
  {
    super.updateInfo(mm);
    if (this.getCountUpdateInfo() > 1) return this;

    // そのステップで受信した(前ステップで送信された)メッセージ一覧
    List<CommunicationMessage> messages =
            mm.getReceivedMessageList();
    this.messageManager = mm;
    return this;

  }

  @Override
  public Search calc()
  {
    EntityID me = this.agentInfo.getID();
    int idx = this.clustering.getClusterIndex(me);
    if (idx == -1) {
      return this;
    }

    Collection<EntityID> cluster =
            this.clustering.getClusterEntityIDs(idx);
    if (cluster == null) {
      return this;
    }

    // 搬送されていない市民を知覚したときメッセージ送信
    List<EntityID> yetLoadCivilians = this.worldInfo.getChanged().getChangedEntities()
            .stream()
            .map(this.worldInfo::getEntity)
            .filter(Civilian.class::isInstance)
            .map(Civilian.class::cast)
            .filter(Civilian::isDamageDefined)
            .filter(e -> e.getDamage()>0) //ダメージあり
            .filter(Civilian::isHPDefined)
            .filter(e -> e.getHP()>0) //死亡していない
            .filter(Civilian::isBuriednessDefined)
            .filter(e -> e.getBuriedness() == 0) //埋没度0
            .map(Civilian::getID)
            .collect(Collectors.toList());

    for(EntityID civilian:yetLoadCivilians){
      Civilian civ = (Civilian) this.worldInfo.getEntity(civilian);
      messageManager.addMessage(new MessageCivilian(false, StandardMessagePriority.HIGH,civ),false);
    }


    // 擬似乱数を用いてランダムに選択するために，Listに変換します
    if (this.result == null){
      list = new ArrayList<>(cluster);
    }

    // エージェントが現在位置する建物や道路のIDを取得する
    EntityID position = this.agentInfo.getPosition();

    // 最初の対象を設定する
    if (this.getTarget() == null) {
      // 擬似乱数を用いてランダムに選択します
      int n = list.size();
      int r = this.random.nextInt(n);
      this.result = list.get(r);

      return this;
    }

    // 対象に辿り着くまで，対象を変更しない
    if (!this.getTarget().equals(position)) {
      return this;
    }

    // 探索済みの (辿り着いたことがある) 対象を候補から除外する
    if (this.getTarget().equals(position)) {
      this.searchedSet.add(this.getTarget());
      list.removeAll(this.searchedSet);

      // すべての対象が探索済みの場合，候補をリセットする
      /*if (list.size() <= 0) {
        this.searchedSet.clear();
        list = new ArrayList<>(cluster);
      }*/

      // 擬似乱数を用いてランダムに選択します
      //int n = list.size();
      //int r = this.random.nextInt(n);
      //this.result = list.get(r);
      //double near = this.worldInfo.getDistance(position, this.list.get(0));
      double near = 999999;
      for (EntityID lists : this.list) {
        StandardEntity test = this.worldInfo.getEntity(lists);
        if(test instanceof Building) {
          double distance = this.worldInfo.getDistance(position, lists);
          if (distance < near){
            near = distance;
            //System.out.println("ok");
            this.result = lists;
          }
        }
      }
      if(this.result.equals(position)){
        this.searchedSet.clear();
        list = new ArrayList<>(cluster);
      }

      return this;
    }

    return this;
  }
}