package AITB_2023.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import org.apache.log4j.Logger;

import AITB_2023.module.util.BuildingStatus;
import AITB_2023.module.util.NextCluster;
import adf.core.agent.action.Action;
import adf.core.agent.action.ambulance.ActionLoad;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.complex.HumanDetector;
import adf.core.debug.DefaultLogger;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class AITATHumanDetector extends HumanDetector {

  private Clustering clustering;

  private EntityID result;

  private Logger logger;

  private Action previousAction = null;
  private HashSet<Human> arleadyRescuedHuman = new HashSet<>();

  private MessageManager messageManager;
  private NextCluster nextCluster;
  private HashSet<EntityID> unsearchedBuildingIDs = new HashSet<>();
  private HashSet<Integer> serachedClusterIndexs = new HashSet<>();
  private int targetClusterIndex;

  public AITATHumanDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager,
      DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    this.clustering = moduleManager.getModule("SampleHumanDetector.Clustering",
        "adf.impl.module.algorithm.KMeansClustering");
    registerModule(this.clustering);
  }

  @Override
  public HumanDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    super.updateInfo(messageManager);
    this.messageManager = messageManager;
    this.UpdateCluster();
    return this;
  }

  private Set<EntityID> getNextClusterEntityIDs(int clusterIndex) {
    HashSet<EntityID> result = new HashSet<>();
    Collection<StandardEntity> clusterEntities = this.clustering.getClusterEntities(clusterIndex);
    clusterEntities.stream()
        .filter(BuildingStatus::isShouldSearch)
        .forEach(entity -> result.add(entity.getID()));
    return result;
  }

  @Override
  public HumanDetector calc() {
    Action action = this.agentInfo.getExecutedAction(this.agentInfo.getTime() - 1);
    if (action instanceof ActionLoad && this.previousAction instanceof ActionLoad) {
      ActionLoad actionLoad = (ActionLoad) action;
      ActionLoad previousActionLoad = (ActionLoad) this.previousAction;
      if (actionLoad.getTarget().equals(previousActionLoad.getTarget())) {
        this.logger.debug("same target");
        Human target = (Human) this.worldInfo.getEntity(actionLoad.getTarget());
        this.arleadyRescuedHuman.add(target);
        this.result = null;
      }
    }

    // 市民を運んでいるかどうか
    Human transportHuman = this.agentInfo.someoneOnBoard();
    if (transportHuman != null) {
      logger.debug("someoneOnBoard:" + transportHuman);
      this.result = transportHuman.getID();
      return this;
    }
    if (this.result != null) {
      Human target = (Human) this.worldInfo.getEntity(this.result);
      if (!isValidHuman(target)) {
        logger.debug("Invalid Human:" + target + " ==>reset target");
        this.result = null;
      }
    }
    // 市民を運んでいなかったらcalcTargetで市民を探す
    if (this.result == null) {
      this.result = calcTarget();
    }

    previousAction = this.agentInfo.getExecutedAction(this.agentInfo.getTime() - 1);
    return this;
  }

  private EntityID calcTarget() {
    HashSet<Human> rescueTargets = filterInCluster(
        filterRescueTargets(this.worldInfo.getEntitiesOfType(StandardEntityURN.CIVILIAN)));
    // HashSet<Human> rescueTargets =
    // filterRescueTargets(this.worldInfo.getEntitiesOfType(StandardEntityURN.CIVILIAN));
    this.logger.debug("rescueTargets:" + rescueTargets);
    if (!rescueTargets.isEmpty()) {
      Human selected = rescueTargets.stream().sorted(new DistanceSorter(this.worldInfo, this.agentInfo.me()))
          .findFirst().get();
      this.logger.debug("Selected:" + selected);
      return selected.getID();
    }
    return null;
  }

  @Override
  public EntityID getTarget() {
    return this.result;
  }

  private HashSet<Human> filterRescueTargets(Collection<StandardEntity> entities) {
    HashSet<Human> rescueTargets = entities.stream()
        .filter(e -> e instanceof Human)
        .filter(e -> isValidHuman((Human) e))
        .map(e -> (Human) e)
        .collect(HashSet::new, HashSet::add, HashSet::addAll);

    rescueTargets.removeAll(this.arleadyRescuedHuman);
    return rescueTargets;
  }

  private HashSet<Human> filterInCluster(HashSet<Human> humans) {
    // int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    // HashSet<StandardEntity> inCluster = new
    // HashSet<>(clustering.getClusterEntities(clusterIndex));
    // return humans.stream()
    // .filter(e -> inCluster.contains(e))
    // .collect(HashSet::new, HashSet::add, HashSet::addAll);
    this.logger.debug("serachedClusterIndexs:" + this.serachedClusterIndexs);
    this.logger.debug("humans:" + humans);
    HashSet<StandardEntity> inCluster = new HashSet<>();
    for (Integer index : this.serachedClusterIndexs) {
      if (index == -1)
        continue;
      inCluster.addAll(this.clustering.getClusterEntities(index));
    }
    return humans.stream()
        .filter(e -> inCluster.contains(this.worldInfo.getEntity(e.getPosition())))
        .collect(HashSet::new, HashSet::add, HashSet::addAll);
  }

  private class DistanceSorter implements Comparator<StandardEntity> {

    private StandardEntity reference;
    private WorldInfo worldInfo;

    DistanceSorter(WorldInfo wi, StandardEntity reference) {
      this.reference = reference;
      this.worldInfo = wi;
    }

    public int compare(StandardEntity a, StandardEntity b) {
      int d1 = this.worldInfo.getDistance(this.reference, a);
      int d2 = this.worldInfo.getDistance(this.reference, b);
      return d1 - d2;
    }
  }

  private boolean isValidHuman(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    if (!target.isDamageDefined() || target.getDamage() == 0)
      return false;
    if (!target.isBuriednessDefined())
      return false;
    if (target.getBuriedness() != 0)
      return false;

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    StandardEntityURN positionURN = position.getStandardURN();
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM)
      return false;

    return true;
  }

  private void UpdateCluster() {
    if (Objects.isNull(this.nextCluster)) {
      this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
          this.clustering.getClusterIndex(this.agentInfo.me()));
    }

    this.unsearchedBuildingIDs.remove(this.worldInfo.getChanged().getChangedEntities());
    if (this.unsearchedBuildingIDs.isEmpty()) {
      this.targetClusterIndex = this.nextCluster.getNextClusterIndex(this.serachedClusterIndexs);
      this.serachedClusterIndexs.add(this.targetClusterIndex);
      if (this.targetClusterIndex == -1) {
        this.unsearchedBuildingIDs.addAll(this.worldInfo.getEntityIDsOfType(StandardEntityURN.BUILDING));
      } else {
        this.unsearchedBuildingIDs.addAll(this.getNextClusterEntityIDs(this.targetClusterIndex));
      }
    }
  }

}