package AITB_2023.module.complex;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import org.apache.log4j.Logger;

import AITB_2023.module.util.BuildingStatus;
import AITB_2023.module.util.NextCluster;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.algorithm.PathPlanning;
import adf.core.component.module.complex.Search;
import adf.core.debug.DefaultLogger;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class AITATSearch extends Search {
  private EntityID result;
  private Clustering clustering;
  private PathPlanning pathPlanning;
  private Logger logger;
  private NextCluster nextCluster;
  private HashSet<EntityID> unsearchedBuildingIDs = new HashSet<>();
  private HashSet<Integer> serachedClusterIndexs = new HashSet<>();
  private int targetClusterIndex;
  private HashMap<EntityID, Integer> patrolBuilding = new HashMap<>();
  private final Integer PATROL_BUILDING_DELAY = 10;
  private HashMap<EntityID, Integer> needHelpCivilian = new HashMap<>();
  private final Integer NEED_HELP_CIVILIAN_SEND_COUNT = 3;

  public AITATSearch(
      AgentInfo ai, WorldInfo wi, ScenarioInfo si,
      ModuleManager mm, DevelopData dd) {
    super(ai, wi, si, mm, dd);

    this.logger = DefaultLogger.getLogger(agentInfo.me());
    this.clustering = mm.getModule("SampleSearch.Clustering.Ambulance",
        "adf.sample.module.algorithm.SampleKMeans");
    this.pathPlanning = moduleManager.getModule(
        "SampleSearch.PathPlanning.Ambulance",
        "adf.impl.module.algorithm.DijkstraPathPlanning");
    this.registerModule(this.clustering);
    this.registerModule(this.pathPlanning);
  }

  @Override
  public EntityID getTarget() {
    return this.result;
  }

  @Override
  public Search updateInfo(MessageManager mm) {
    super.updateInfo(mm);
    if (this.getCountUpdateInfo() > 1)
      return this;

    if (Objects.isNull(this.nextCluster)) {
      this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
          this.clustering.getClusterIndex(this.agentInfo.me()));
    }

    this.unsearchedBuildingIDs.remove(this.worldInfo.getChanged().getChangedEntities());
    if (this.unsearchedBuildingIDs.isEmpty()) {
      this.targetClusterIndex = this.nextCluster.getNextClusterIndex(this.serachedClusterIndexs);
      this.serachedClusterIndexs.add(this.targetClusterIndex);
      if (this.targetClusterIndex == -1) {
        this.unsearchedBuildingIDs.addAll(this.worldInfo.getEntityIDsOfType(StandardEntityURN.BUILDING));
      } else {
        this.unsearchedBuildingIDs.addAll(this.getNextClusterEntityIDs(this.targetClusterIndex));
      }
    }

    for (EntityID entityID : this.worldInfo.getChanged().getChangedEntities()) {
      StandardEntity entity = this.worldInfo.getEntity(entityID);
      if (entity instanceof Human) {
        Human human = (Human) entity;
        if (!human.isHPDefined() || human.getHP() == 0)
          continue;
        if (human.isBuriednessDefined() && human.getBuriedness() == 0)
          continue;
        if (!human.isPositionDefined())
          continue;
        if (this.patrolBuilding.containsKey(human.getPosition()))
          continue;
        this.patrolBuilding.put(human.getPosition(), human.getBuriedness() + this.PATROL_BUILDING_DELAY);
        this.needHelpCivilian.put(human.getID(), this.NEED_HELP_CIVILIAN_SEND_COUNT);
      }
    }

    HashSet<EntityID> removeEntityIDs = new HashSet<>();
    for (EntityID entityID : this.patrolBuilding.keySet()) {
      if (this.patrolBuilding.get(entityID) < 0)
        removeEntityIDs.add(entityID);
      this.patrolBuilding.put(entityID, this.patrolBuilding.get(entityID) - 1);
    }

    for (EntityID entityID : removeEntityIDs) {
      this.patrolBuilding.remove(entityID);
    }

    // need help civilianから一つランダムに選んで、その建物を通信する
    if (this.needHelpCivilian.size() > 0) {
      EntityID entityID = (EntityID) this.needHelpCivilian.keySet().toArray()[0];
      StandardEntity entity = this.worldInfo.getEntity(entityID);
      if (entity instanceof Civilian) {
        Civilian civilian = (Civilian) entity;
        mm.addMessage(new MessageCivilian(false, civilian));
        this.needHelpCivilian.put(entityID, this.needHelpCivilian.get(entityID) - 1);
      }
    }

    for (EntityID entityID : this.needHelpCivilian.keySet()) {
      StandardEntity entity = this.worldInfo.getEntity(entityID);
      if (entity instanceof Civilian) {
        Civilian civilian = (Civilian) entity;
        mm.addMessage(new MessageCivilian(true, civilian));
      }
    }

    // need help civilianから0以下のものを削除
    removeEntityIDs = new HashSet<>();
    for (EntityID entityID : this.needHelpCivilian.keySet()) {
      if (this.needHelpCivilian.get(entityID) < 0)
        removeEntityIDs.add(entityID);
    }

    for (EntityID entityID : removeEntityIDs) {
      this.needHelpCivilian.remove(entityID);
    }

    return this;
  }

  @Override
  public Search calc() {
    this.result = null;
    logger.debug("unsearchedBuildingIDs: " + this.unsearchedBuildingIDs);
    logger.debug("targetClusterIndex: " + this.targetClusterIndex);

    for (EntityID entityID : this.patrolBuilding.keySet()) {
      if (this.patrolBuilding.get(entityID) == 0) {
        this.logger.debug("patrolBuilding: " + entityID);
        this.result = entityID;
        return this;
      }
    }

    this.result = this.unsearchedBuildingIDs.stream()
        .sorted((id1, id2) -> {
          double distance1 = this.worldInfo.getDistance(this.agentInfo.me().getID(), id1);
          double distance2 = this.worldInfo.getDistance(this.agentInfo.me().getID(), id2);
          return Double.compare(distance1, distance2);
        })
        .findFirst()
        .orElse(null);

    return this;
  }

  private Set<EntityID> getNextClusterEntityIDs(int clusterIndex) {
    HashSet<EntityID> result = new HashSet<>();
    Collection<StandardEntity> clusterEntities = this.clustering.getClusterEntities(clusterIndex);
    clusterEntities.stream()
        .filter(BuildingStatus::isShouldSearch)
        .forEach(entity -> result.add(entity.getID()));
    return result;
  }
}