package AITB_2023.module.complex;

import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.communication.standard.bundle.information.MessageRoad;
import adf.core.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.core.agent.info.*;
import adf.core.component.module.complex.HumanDetector;
import adf.core.component.communication.CommunicationMessage;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import java.util.stream.Stream;
import java.util.stream.Collectors;
import adf.core.component.module.algorithm.Clustering;
import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;
import AITB_2023.module.util.NextCluster;

import static adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam.ACTION_LOAD;
import static java.lang.Double.POSITIVE_INFINITY;
import static java.lang.Integer.MAX_VALUE;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;

//timeがステップ数 targetが市民のエンティティID 
// 行動(掘り起こし)対象(市民)を決定するクラス
public class AITFBHumanDetector extends HumanDetector
{
    private EntityID result;
    private Clustering clustering;
    private MessageManager messageManager;
    private Logger logger;

    private Collection<EntityID> alreadyDigupIDs;// 既に掘り起こしたエージェントのIDのコレクション

    // 1step前の情報を保持するフィールド群
    private EntityID lastID = null;
    private double lastX = -1.0;//1step前のエージェント自身のX座標
    private double lastY = -1.0;//1step前のエージェント自身のY座標
    private int lastTargetBuriedness = -1;//1step前の埋没度

    //通信用のアクション番号
    final private int HELP_BLOCKADE = 6;//瓦礫で動けない状態

    //エージェントごとの掘り起こし優先度
    private int civ_priority = 2;
    private int at_priority = 1;
    private int fb_priority = 1;
    private int pf_priority = 1;

    // クラスタ関連
    private NextCluster nextCluster;
    private int myClusterIdx;
    private Collection<StandardEntity> myClusterEntities;
    private HashSet<Integer> searchedClusterIndexs;

    //ボイスメッセージ関連
    private Civilian civ;
    private int messageStep = 1;
    private boolean rescueMessage = false;

    public AITFBHumanDetector(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);
        this.clustering = mm.getModule("SampleHumanDetector.Clustering", "adf.impl.module.algorithm.KMeansClustering");
        this.registerModule(this.clustering);

        this.logger = DefaultLogger.getLogger(agentInfo.me());

        this.searchedClusterIndexs = new HashSet<>();
        this.alreadyDigupIDs = new ArrayList<>();// 初期化
    }

    //行動対象をreturnするメソッド
    @Override
    public EntityID getTarget()
    {
        return this.result;
    }
    
    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        // 別のモジュールでregisterModuleを書いたとき、２回目以降は実行されなくなる
        if (this.getCountUpdateInfo() > 1) return this;
        logger.debug("time: "+this.agentInfo.getTime()+" Detector_updateInfo");
        
        this.messageManager = messageManager;
        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages = messageManager.getReceivedMessageList(MessageCivilian.class);

        // 1stepに担当するクラスタの番号とクラスタ内のエンティティコレクションを取得
        if (this.agentInfo.getTime() == 1){
            // nextClusterを設定する
            this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
                this.clustering.getClusterIndex(this.agentInfo.me()));
            //このエージェントが担当するクラスタの番号を取得し，myClusterIdxに代入
            this.myClusterIdx = this.clustering.getClusterIndex(this.agentInfo.getID());
            //idx番のクラスタに属するEntityをすべて取得し，myClusterEntitiesに代入
            this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
        }

        this.result = null;// 現在の掘り起こし対象リセットする

        if(this.agentInfo.getTime() > 3){// 最初の3stepは消防隊が止まってるから3stepまでは行わない
            // 自分が瓦礫で止まっているかを判定して真なら土木隊に送信
            this.message_police();
        }
        
        // 架空のエージェントを掘り起こしていたなら
        if(isAirAgent()){
            this.alreadyDigupIDs.add(this.lastID);// alreadyDigupIDsに追加
            this.lastID = null;
        }
        
        // 1step前の救助対象のIDからentityを取得
        StandardEntity lastEntity = this.worldInfo.getEntity(lastID);
        // lastEntityを掘り起こし終わった ならば
        if(isDigupfinish()){
            this.alreadyDigupIDs.add(this.lastID);// 掘り起こし終わったIDをalreadyDigupIDsに追加
            
            if(lastEntity instanceof Civilian){// lastEntityインスタンスがCivilianクラスなら

                civ = (Civilian) lastEntity;//ダウンキャスト
                //this.logger.debug("市民" +lastID+ "を掘り起こした" +" resultは"+this.result);
                messageStep = this.agentInfo.getTime() + 2;
                rescueMessage = true;
                //plt_mess(" 市民を掘り起こした");
                //messageManager.addMessage(new MessageCivilian(false,civ),false);//救助した市民情報を送信

            }
        }

        //市民を掘り起こした後，3stepの間市民を救出したボイスメッセージを送信する
        if(rescueMessage){
            plt_mess(" 市民を掘り起こした");
            messageManager.addMessage(new MessageCivilian(false,civ),false);//救助した市民情報を送信
            if (this.agentInfo.getTime() == messageStep){
                rescueMessage = false;
            }
        }

        // stepが60より大きい && 現在掘り起こしていない ならば
        if(this.agentInfo.getTime()>60 && !isDigup()){
            changeDigupPriority();// 掘り起こし優先度の設定を変更
        }

        return this;
    }

    // 行動対象を計算するメソッド
    @Override
    public HumanDetector calc()
    {
        logger.debug("time: "+this.agentInfo.getTime()+" Detector_calc");

        // 市民を掘り起こしているなら
        if(isDigup()){
            this.result = this.lastID;// 掘り起こし対象を継続する
            return this;
        }

        // 救助対象を探索
        EntityID targetID = this.detectDigupTarget();
        this.result = targetID;

        this.lastID = this.result;// 最後に現在の救助対象IDをlastIDとして格納
        this.lastX = this.agentInfo.getX();// 今のXをlastXに格納
        this.lastY = this.agentInfo.getY();// 今のYをlastYに格納
        // 今の埋没度をlastTargetBuriednessに格納
        if(this.result!=null){this.lastTargetBuriedness=((Human)this.worldInfo.getEntity(this.result)).getBuriedness();}else{this.lastTargetBuriedness=-1;}

        return this;
    }

    // 救助対象を検出する
    public EntityID detectDigupTarget(){
        // 知覚している市民,消防隊,救急隊,土木隊のEntityIDコレクションを取得
        Collection<EntityID> ids = this.worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN,AMBULANCE_TEAM,FIRE_BRIGADE,POLICE_FORCE);
        // 既に掘り起こしたエージェントのIDをidsから消しておく
        ids.removeAll(this.alreadyDigupIDs);

        // 探索クラスタIDセットから探索クラスタ内のEntityリストを生成
        Collection<StandardEntity> inCluster = new HashSet<>();
        if(this.myClusterIdx!=-1)inCluster.addAll(this.myClusterEntities);
        for(Integer i : this.searchedClusterIndexs){
            inCluster.addAll(this.clustering.getClusterEntities(i.intValue()));
        }

        // 救助対象を探索して救助対象リストを生成
        List<EntityID> diguptarget_list = ids.stream()
        .map(this.worldInfo::getEntity)
        .map(entity -> (Human)entity)
        // 場所のEntityが探索クラスタEntityリストに含まれているエージェントを抽出
        .filter(hum -> inCluster.contains(this.worldInfo.getEntity(hum.getPosition())) )
        // 生きているエージェントを抽出
        .filter(hum -> hum.isHPDefined() && hum.getHP() > 0)
        // 埋まっているエージェントを抽出
        .filter(hum -> hum.isBuriednessDefined() && hum.getBuriedness() > 0)
        // 掘り起こし優先度が0以外のエージェントを抽出
        .filter(hum -> this.getDigupPriority(hum)!=0)
        // 掘り起こし優先度の昇順で，同一優先度の間は自身との距離の昇順となるようにソート
        .sorted(Comparator.comparing(this::getDigupPriority).thenComparing(hum -> this.worldInfo.getDistance(this.agentInfo.getID(), hum.getID())))
        .map(hum -> hum.getID())
        .collect(Collectors.toList());// 救助対象リストを生成

        Optional<EntityID> diguptarget = diguptarget_list.stream().findFirst();// streamの先頭要素オブジェクトを取得
        EntityID targetID = (diguptarget.isPresent() ? diguptarget.get() : null);
        //plt_mess("\n"+diguptarget_list+"\nOptional:"+targetID);
        return targetID;
    }
    
    // 現在、瓦礫で身動きが取れないかを判断し、真なら土木隊に送信する処理
    private void message_police(){
        FireBrigade fbMeEntity = (FireBrigade)this.agentInfo.me();// 自身のEntity(FireBrigade)
        EntityID nowPositionID = this.agentInfo.getPosition();// 現在場所のID
        
        // エージェントが瓦礫で止まっているとき
        if(isStop_broken()){
            this.logger.debug("現在、Road居てさっきとエージェントの場所が変わってないので瓦礫で止まってます");
            if(this.lastID==null){this.plt_mess(" 瓦礫で止まってます");}

            // 消防隊が埋まったという情報を送信
            messageManager.addMessage(new MessageFireBrigade(false,fbMeEntity,HELP_BLOCKADE,nowPositionID),false);

        }
    }
    // エージェントが瓦礫で止まっているかを判定するメソッド
    private boolean isStop_broken(){
        EntityID nowPositionID = this.agentInfo.getPosition();//エージェント自身が現在居る場所のID
        StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);//エージェント自身が現在居るエリアのIDからエリアのエンティティを取得
        double[] pos_dif = compDistanceXY();//エージェント自身の現在の位置と1step前の位置の差を取得{x,y}

        // もしエージェントが現在いる場所が道路で、1step前と場所が全く変わっていないとき
        if((nowPositionEntity instanceof Road) && ((pos_dif[0] < 3.0)&&(pos_dif[1] < 3.0))){
            return true;
        }else{
            return false;
        }
    }

    // 掘り起こしているかを判定するメソッド
    private boolean isDigup(){
        EntityID nowPositionID = this.agentInfo.getPosition();// エージェント自身の現在場所のID
        StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);// エージェント自身が現在居る場所のエンティティをIDから取得
        double[] pos_dif = compDistanceXY();// エージェント自身の現在の位置と1step前の位置の差を取得{x,y}

        if(this.lastID==null) return false;// 1step前の値がnull->掘り起こしていない

        Human lastEntity = (Human)this.worldInfo.getEntity(this.lastID);//1step前の救助対象を取得
        // もしエージェントが現在いる場所が建物で、1step前と場所が全く変わっておらず、lastEntityの埋没度が0より大きいとき
        if((nowPositionEntity instanceof Building) && ((pos_dif[0] < 3.0)&&(pos_dif[1] < 3.0)) && lastEntity.getBuriedness() > 0){
            return true;
        }else{
            return false;
        }
    }
    // 掘り起こし対象から掘り起こし終わったかを判定するメソッド
    private boolean isDigupfinish(){
        // 1step前の救助対象IDがnull->掘り起こし対象が対象がいなかったのでfalse
        if(lastID == null) return false;

        Human lastEntity = (Human)this.worldInfo.getEntity(this.lastID);//1step前の救助対象を取得
        // 1step前の救助対象の埋没度が0
        if(lastEntity.getBuriedness() == 0){
            return true;
        }else{
            return false;
        }
    }

    // 1step前の救助対象が架空のエージェントかを判定するメソッド
    private boolean isAirAgent(){
        // 1step前の値がnull->救助対象がいない
        if(this.lastID == null) return false;

        Human lastEntity = (Human)this.worldInfo.getEntity(this.lastID);
        EntityID lastEntity_posi = lastEntity.getPosition();
        
        // step前の救助対象と自身の位置を比較->異なるなら架空かわからないのでfalse
        if(!lastEntity_posi.equals(this.agentInfo.getPosition())) return false;

        Collection<EntityID> changes = this.worldInfo.getChanged().getChangedEntities();// 視覚内のエージェントのIDを取得
        // 1step前の救助対象のIDがchangesに含まれているなら
        if(!changes.contains(this.lastID)){
            plt_mess(" 救助対象は架空のエージェントだった!");
            logger.debug("Time: "+this.agentInfo.getTime()+" 救助対象は架空のエージェントだった!");
            return true; // 救助対象を架空のエージェントにしていた
        }else{
            return false;
        }
    }

    // エージェント自身の現在の位置と1step前の位置の差を出力するメソッド
    private double[] compDistanceXY(){
        double nowX = this.agentInfo.getX();//現在のX座標を取得
        double nowY = this.agentInfo.getY();//現在のY座標を取得
        double posX_dif = Math.abs((nowX - this.lastX));//現在のX座標と1step前のX座標との差
        double posY_dif = Math.abs((nowY - this.lastY));//現在のY座標と1step前のY座標との差
        double[] pos_dif = {posX_dif,posY_dif};//X座標の差とY座標の差を持つ配列
        return pos_dif;
    }

    // 掘り起こし優先度を取得するメソッド
    private int getDigupPriority(Human hum){
        StandardEntity entity = hum;
        int digup_priority = 0;// Humanの掘り起こし優先度を格納する変数
        //entityに紐ずく掘り起こし優先度を取得
        switch(entity.getStandardURN()){
            case CIVILIAN://entityが市民であれば
                digup_priority = this.civ_priority;
                break;
            case AMBULANCE_TEAM://entityが救助隊であれば
                digup_priority = this.at_priority;
                break;
            case FIRE_BRIGADE://entityが消防隊であれば
                digup_priority = this.fb_priority;
                break;
            case POLICE_FORCE://entityが土木隊であれば
                digup_priority = this.pf_priority;
                break;
        }
        return digup_priority;
    }

    // 掘り起こし優先度の設定を変更するメソッド
    private void changeDigupPriority(){
        this.civ_priority = 1;
        this.at_priority = 0;
        this.fb_priority = 0;
        this.pf_priority = 0;
        //logger.debug("掘り起こし優先度を変更しました");
    }

    // 担当クラスタを変更するメソッド
    public void change_cluster(int idx, HashSet<Integer> scIndexs){
        this.myClusterIdx = idx;
        this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
        this.searchedClusterIndexs = new HashSet<>(scIndexs);
        plt_mess(" クラスタを変更しました idx="+this.myClusterIdx+"\n"+this.searchedClusterIndexs);
    }

    public void plt_mess(String mess){
        if(this.lastID == null){
            System.out.println("FB_D> time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+" 対象ID: "+"null"+mess);
            logger.debug("time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+" 対象ID: "+"null"+mess);
        }else{
            System.out.println("FB_D> time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+" 対象ID: "+this.lastID.getValue()+mess);
            logger.debug("time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+" 対象ID: "+this.lastID.getValue()+mess);
        }
    }
}
