package AITB_2023.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_CENTRE;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_OFFICE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.StandardMessagePriority;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.info.*;
import adf.core.agent.precompute.PrecomputeData;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.complex.Search;
import adf.core.component.module.complex.HumanDetector;
import AITB_2023.module.complex.AITFBHumanDetector;
import adf.core.component.module.algorithm.Clustering;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import java.util.stream.Collectors;
import AITB_2023.module.util.NextCluster;

import static java.lang.Double.POSITIVE_INFINITY;

import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;

public class AITFBSearch extends Search
{
    private EntityID result;
    private Random random = new Random(1);
    private Clustering clustering;
    private List<EntityID> unsearchedBuildingIDs;
    private MessageManager messageManager;
    private HumanDetector humanDetector;
    private static final int ACTION_LOAD = CommandAmbulance.ACTION_LOAD;
    private Logger logger;

    // クラスタ関連
    private NextCluster nextCluster;
    private int myClusterIdx;
    private Collection<StandardEntity> myClusterEntities;
    private HashSet<Integer> searchedClusterIndexs;

    public AITFBSearch(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);
        this.logger = DefaultLogger.getLogger(agentInfo.me());

        // 利用するクラスタリングモジュールの読み込み
        String clusteringKey = "SampleSearch.Clustering";
        switch (ai.me().getStandardURN())
        {
            case FIRE_BRIGADE:
                clusteringKey += ".Fire";
                break;
            case AMBULANCE_TEAM:
                clusteringKey += ".Ambulance";
                break;
            case POLICE_FORCE:
                clusteringKey += ".Police";
                break;
        }

        this.clustering =
                mm.getModule(clusteringKey,
                        "adf.sample.module.algorithm.SampleKMeans");
        this.registerModule(this.clustering);

        // Detectorを読み込む
        this.humanDetector = mm.getModule("DefaultTacticsFireBrigade.HumanDetector","AITB_2023.module.complex.AITFBHumanDetector");
        this.registerModule(this.humanDetector);

        this.unsearchedBuildingIDs = new ArrayList<>();
        this.searchedClusterIndexs = new HashSet<>();
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    @Override
    public Search updateInfo(MessageManager mm)
    {
        super.updateInfo(mm);
        // 別のモジュールでregisterModuleを書いたとき、２回目以降は実行されなくなる
        if (this.getCountUpdateInfo() > 1) return this;

        logger.debug("time: "+this.agentInfo.getTime()+" Search_updateInfo");
        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages =
                mm.getReceivedMessageList();
        this.messageManager = mm;

        /*List<StandardEntity> a = this.worldInfo.getChanged().getChangedEntities().stream()
        .map(this.worldInfo::getEntity)
        .collect(Collectors.toList());
        List<StandardEntity> b =  a.stream()
        .filter(entity -> entity instanceof Building)
        .collect(Collectors.toList());
        plt_mess("\n"+a+"\n"+b);*/
        
        EntityID position = this.agentInfo.getPosition();// エージェントが現在位置する建物や道路のIDを取得する
        this.unsearchedBuildingIDs.remove(position);

        // もし未探索IDコレクションに要素がなかったら
        if (this.unsearchedBuildingIDs.isEmpty()) {
            this.reset();// リセットする
            this.unsearchedBuildingIDs.remove(this.agentInfo.getPosition());
        }

        return this;
    }

    @Override
    public Search calc()
    {
        logger.debug("time: "+this.agentInfo.getTime()+" Search_calc");

        // 搬送されていない市民を知覚したときメッセージ送信
        yetLoadCivilian_send();

        //エージェント自身のEntityIDを取得
        EntityID me = this.agentInfo.getID();
        // エージェントが現在位置する建物や道路のIDを取得する
        EntityID position = this.agentInfo.getPosition();

        // 最初の対象をランダムに選択する
        if (this.getTarget() == null) {
            
            // 擬似乱数を用いてランダムに選択します
            int n = this.unsearchedBuildingIDs.size();
            if(n>0) {
                int r = this.random.nextInt(n);
                this.result = this.unsearchedBuildingIDs.get(r);
            }

            return this;
        }
        String str = " lastID: "+this.getTarget();
        plt_mess(str);

        // 対象に辿り着くまで，対象を変更しない
        if (this.unsearchedBuildingIDs.contains(this.getTarget()) ){
            return this;
        }

        double near = Double.POSITIVE_INFINITY;//nearesDistanceに初期値としてDoubleで最も大きい値を代入
        // 一番距離が近い建物を探索
        for (EntityID id : this.unsearchedBuildingIDs) {
            StandardEntity test = this.worldInfo.getEntity(id);
            if(test instanceof Building) {
                double distance = this.worldInfo.getDistance(me, id);
                if (distance < near){
                    near = distance;
                    this.result = id;
                }
            }
        }
        plt_mess(" this.result: "+this.getTarget());
        return this;
    }

    // 探索コレクションをリセットするメソッド
    private void reset() {
        this.unsearchedBuildingIDs.clear();//探索していない建物IDコレクションの要素を全て消す
        // 担当クラスタを決定
        if (this.agentInfo.getTime() == 1){//もしstepが1なら
            // nextClusterを設定する
            this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
                this.clustering.getClusterIndex(this.agentInfo.me()));
            // 自分が担当するクラスタのクラスタIDを取得
            this.myClusterIdx = this.clustering.getClusterIndex(this.agentInfo.getID());
            // クラスタIDからクラスタ内のエンティティを取得
            this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
            plt_mess(" クラスタ初期設定: \n"+this.myClusterIdx+" "+this.myClusterEntities);
        }else{
            change_cluster();//担当クラスタを変更する処理
        }
        // クラスタ内エンティティコレクションがnullでなく、要素数が0以上のとき
        if (this.myClusterEntities != null && this.myClusterEntities.size() > 0) {
            for (StandardEntity entity : this.myClusterEntities) {//拡張for
                // もしエンティティがBuildingクラスかつEntityのenumがREFUGEでないなら
                if (entity instanceof Building && entity.getStandardURN() != REFUGE) {
                    this.unsearchedBuildingIDs.add(entity.getID());// 未探索コレクションにEntityIDを追加
                }
            }
        } else {
            this.unsearchedBuildingIDs
                    .addAll(this.worldInfo.getEntityIDsOfType(BUILDING));
                    /*.addAll(this.worldInfo.getEntityIDsOfType(BUILDING, GAS_STATION,
                            AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE));*/
            plt_mess(" targetクラスタなし unsearchedBuildingIDs: "+this.unsearchedBuildingIDs);
        }
    }

    // 搬送されていない市民を知覚したときメッセージを送信するメソッド
    private void yetLoadCivilian_send(){
        List<EntityID> yetLoadCivilians = this.worldInfo.getChanged().getChangedEntities()
                .stream()
                .map(this.worldInfo::getEntity)
                .filter(Civilian.class::isInstance)
                .map(Civilian.class::cast)
                .filter(Civilian::isDamageDefined)
                .filter(e -> e.getDamage()>0) //ダメージあり
                .filter(Civilian::isHPDefined)
                .filter(e -> e.getHP()>0) //死亡していない
                .filter(Civilian::isBuriednessDefined)
                .filter(e -> e.getBuriedness() == 0) //埋没度0
                .map(Civilian::getID)
                .collect(Collectors.toList());

        // メッセージの送信
        for(EntityID civilian:yetLoadCivilians){
            Civilian civ = (Civilian) this.worldInfo.getEntity(civilian);
            messageManager.addMessage(new MessageCivilian(false, StandardMessagePriority.HIGH,civ),false);
        }
    }
    
    // 担当クラスタを変更するメソッド
    private void change_cluster(){
        AITFBHumanDetector aitFbHumanDetector = (AITFBHumanDetector)this.humanDetector;
        if(aitFbHumanDetector.detectDigupTarget() != null) return;//detectorが持つクラスタにまだ救助対象があればクラスタを変更しない
        
        // 探索済みクラスタリストに次のクラスタIDを追加
        if(this.myClusterIdx!=-1) this.searchedClusterIndexs.add(this.myClusterIdx);// -1を追加するとDetectorでバグる
        // 次のクラスタのIDを取得
        this.myClusterIdx = this.nextCluster.getNextClusterIndex(this.searchedClusterIndexs);
        this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);

        // 消防隊humanDetectorの方も変える
        aitFbHumanDetector.change_cluster(this.myClusterIdx,this.searchedClusterIndexs);
        plt_mess(" クラスタ距離 \n"+this.nextCluster.getStr());
        plt_mess(" クラスタ変更 "+" idx: "+ this.myClusterIdx + " entities: " + this.myClusterEntities);
    }

    public void plt_mess(String mess){
        System.out.println("FB_S> "+"time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+mess);
        logger.debug("time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+mess);
    }
}
