package AITB_2023.module.complex;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import static rescuecore2.standard.entities.StandardEntityURN.ROAD;
import adf.core.agent.info.*;
import adf.core.component.module.complex.RoadDetector;
import adf.core.component.module.complex.Search;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.algorithm.Clustering;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import java.util.stream.Collectors;

import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;

/*
 * 作成したモジュールの読み込み
 */
import AITB_2023.module.complex.AITRoadDetector;
import AITB_2023.module.util.NextCluster;

/*
 * memo欄
 */

public class AITPFSearch extends Search 
{
    /*
     * result => 探索対象
     * clustering => 
     * list => 探索していないEntityID
     * logger => 
     */
    private EntityID result;
    private Clustering clustering;
    private List<EntityID> list;
    private Logger logger;
    private Random random = new Random(1);
    /*
     * 通信用変数定義（予定）
     * 
     */
    private MessageManager messageManager;
    private List<CommunicationMessage> messages;
    private CommunicationMessage message;
    private Set<Civilian> changedCivilians;
    private List<CommunicationMessage> firstmessages;
    private List<CommunicationMessage> secondmessages;
    /*
     * クラスタ変更用関数定義（予定）
     * 
     */
    private RoadDetector roadDetector;
    private NextCluster nextCluster;
    private int myClusterIdx;
    private Collection<StandardEntity> myClusterEntities;
    private int firstClusterIdx;
    private Collection<StandardEntity> firstClusterEntities;
    private HashSet<Integer> searchedClusterIndexs;

    public AITPFSearch(
        AgentInfo ai, WorldInfo wi, ScenarioInfo si,
        ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);
        this.logger = DefaultLogger.getLogger(agentInfo.me());

        /*
         * クラスタリングモジュールの読み込み
         */
        String clusteringKey = "SampleSearch.Clustering.Police";
        this.clustering =
            mm.getModule(clusteringKey,
                    "adf.sample.module.algorithm.SampleKMeans");
        this.registerModule(this.clustering);
        /*
         * AITRoadDetectorの読み込み
         */
        this.roadDetector = mm.getModule("DefaultTacticsPoliceForce.RoadDetector","AITB_2023.module.complex.AITRoadDetector");
        this.registerModule(this.roadDetector);
        /*
         * インスタンス生成
         */
        this.list = new ArrayList<>();
        this.firstmessages = new ArrayList<>();
        this.secondmessages = new ArrayList<>();
        this.changedCivilians = new HashSet<>();
        this.searchedClusterIndexs = new HashSet<>();
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    /*
     * calcの前に動く
     * Detecterのcalcで止まって欲しくない処理はここ
     */
    @Override
    public Search updateInfo(MessageManager mm)
    {
        super.updateInfo(mm);

        /*
         * エージェントの現在位置のEntitiyID
         */
        EntityID position = this.agentInfo.getPosition();
        /*
         * 別のモジュールでregisterModuleを書いた時，2回目以降は実行されなくなる
         */
        if (this.getCountUpdateInfo() > 1) return this;
        /*
         * 知覚した市民の情報を送信
         * この場合，歩いている市民とかも送ってるため，変更を検討中
         */
        this.two_times_communication(mm);
        /*
         * 現在位置をリストから削除
         */
        if(!this.list.isEmpty()) this.list.remove(position);
        /*
         * 探索し終わったらクラスタを移動
         */
        if(this.list.isEmpty()) 
        {
            this.reset_cluster();
            this.list.remove(position);
        }

        return this;
    }

    @Override
    public Search calc()
    {
        /*
         * position => 現在位置のEntityId
         */
        EntityID position = this.agentInfo.getPosition();
        /*
         * 探索対象がない場合，対象を決定
         */
        if(this.result == null) 
        {
            this.result = this.targetset(position);  
            return this;
        }
        /*
         * 目的地が現在地になっていない時，そのままの目的地にしておく
         * 目的地が現在地の時,listが１以下であればその場所を，それ以外であれば近い場所を探索対象に
         */
        if (!this.getTarget().equals(position))
        {
            return this;
        }else{
            if(this.list.size() <= 1)
            {
                this.list.clear();
                this.result = position;
            }else{
                this.list.remove(position);
                this.result = targetset(position);
            }
        }

        return this;
    }

    /*
     * 探索対象の決定
     * 近い道路から対象を選択
     * ps => 自分の場所
     * rs => 探索対象
     */
    private EntityID targetset(EntityID ps)
    {
        EntityID me = this.agentInfo.getID();
        EntityID rs = null;
        /*
         * listに二つ以上の要素がある時，近いものを選択
         */
        if(this.list.size() > 1) 
        {
            //書き換え中
            double near = Double.POSITIVE_INFINITY;
            for(EntityID ls : this.list)
            {
                double distance = this.worldInfo.getDistance(me, ls);
                if (distance < near){
                    near = distance;
                    rs = ls;
                }
            }
        }
        /*
         * rsがnullになった場合，とりあえずランダムに設定
         */
        if(rs == null)
        {
            int n = this.list.size();
            int r = this.random.nextInt(n);
            rs = this.list.get(r);
        }

        return rs;
    }

    /*
     * 通信関係の処理
     * messages => そのステップで受信した(前ステップで送信された)メッセージ一覧
     * messageManager => thisで使えるようにしてる
     */
    private void communication_processing(MessageManager mm)
    {
        this.messages = mm.getReceivedMessageList();
        this.messageManager = mm;
    }

    /*
     * 同じ通信を2回続ける処理
     * 
     */
    private void two_times_communication(MessageManager mm)
    {
        /* 
         * 知覚範囲から市民を抽出
         * changed => そのステップで知覚したエンティティ(のID)一覧
         * ↑市民のみを対象とするため使用せず
         * changedCivilians => そのステップで知覚した市民(のID)一覧
        */
        this.changedCivilians = this.worldInfo.getChanged().getChangedEntities()
            .stream()
            .map(this.worldInfo::getEntity)
            .filter(Civilian.class::isInstance)
            .map(Civilian.class::cast)
            .collect(Collectors.toSet());        
        /*
         * 前回保存したメッセージを移動
         * firstmessages => 前回のメッセージSet
         * secondmessages => 移動先のメッセージSet
         */
        if (!firstmessages.isEmpty()) 
        {
            this.secondmessages = this.firstmessages;
            this.firstmessages.clear();
        }        
        /*
         * メッセージを作成
         * 作成したメッセージを送信，firstmessagesに保存
         * 現状：通信手段は考慮してない
         */
        if(this.changedCivilians != null)
        {
            for(Civilian cv: changedCivilians) 
            {
                this.message = new MessageCivilian(false, cv);
                mm.addMessage(this.message);
                this.firstmessages.add(this.message);
            }
        }
        /*
         * 前回のメッセージを再度送信
         */
        if(this.secondmessages != null)
        {
            for(CommunicationMessage cm: secondmessages)
            {
                mm.addMessage(cm);
            }
        }        
        this.communication_processing(mm);
    }

    /*
     * クラスタ内の情報をリセット
     */
    private void reset_cluster()
    {
        /*
         * 探索していないリスト削除
         */
        this.list.clear();
        /*
         * 初回はnextClusterを設定
         * それ以外はクラスタ変更
         * nextCluster =>
         * myClusterIdx => 自分が担当するクラスタのクラスタID
         * myClusterEntities => 担当クラスタ内のエンティティ
         */
        if (this.agentInfo.getTime() == 1)
        {
            this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
                this.clustering.getClusterIndex(this.agentInfo.me()));
            this.myClusterIdx = this.clustering.getClusterIndex(this.agentInfo.getID());
            this.firstClusterIdx = this.myClusterIdx;
            this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
            this.firstClusterEntities = this.myClusterEntities;
        }else{
            this.change_cluster();
        }
        /*
         * クラスタ内エンティティコレクションがnullの時，初期クラスタに戻る
         */
        if(this.myClusterEntities == null || this.myClusterEntities.isEmpty())
        {
            this.myClusterIdx = this.firstClusterIdx;
            this.myClusterEntities = this.firstClusterEntities;
        }
        /*
         * クラスタ内エンティティコレクションがnullでなく、要素数が0以上のとき
         * listを再設定
         * それ以外は道路を全てlistに
         */
        if (this.myClusterEntities != null && !this.myClusterEntities.isEmpty())
        {
            for (StandardEntity entity : this.myClusterEntities) {
                if (entity instanceof Road) {
                    this.list.add(entity.getID());
                }
            }
        } else {
            this.list.addAll(this.worldInfo.getEntityIDsOfType(ROAD));
        }
    }

    /*
     * クラスタ移動の処理
     */
    private void change_cluster()
    {
        /*
         * roadDetectorをAITRoadDetectorにキャスト
         * detectorが持つクラスタにまだ活動対象があればクラスタを変更しない
         * ↑ディテクター待ち
         * aitRoaddetector => AITRoadDetectorを使えるようにしてる
         */
        AITRoadDetector aitRoadDetector = (AITRoadDetector)this.roadDetector;
        if(aitRoadDetector.getTarget() != null) return;
        /*
         * 探索済みクラスタリストに次のクラスタIDを追加
         * -1を追加するとDetectorでバグる
         */
        if(this.myClusterIdx != -1) this.searchedClusterIndexs.add(this.myClusterIdx);
        /*
         * 次のクラスタのIDを取得
         * 探索クラスタリストに次のクラスタIDを追加
         * myClusterIdx => 次のクラスタのID
         * myClusterEntities => 次のクラスタのエンティティの集合
         */
        this.myClusterIdx = this.nextCluster.getNextClusterIndex(this.searchedClusterIndexs);
        this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
        /*
         * AITRoadDetectorも同期する
         */
        aitRoadDetector.change_cluster(this.myClusterIdx,this.searchedClusterIndexs);
    }
}

/*
            rs = this.list.stream()
                .sorted((id1, id2) -> {
                double distance1 = this.worldInfo.getDistance(ps, id1);
                double distance2 = this.worldInfo.getDistance(ps, id2);
                return Double.compare(distance1, distance2);
                })
                .findFirst()
                .orElse(null);
            */