package AITB_2023.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.core.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.communication.standard.bundle.centralized.*;
import adf.core.agent.communication.standard.bundle.information.*;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.algorithm.PathPlanning;
import adf.core.component.module.complex.RoadDetector;
import adf.core.debug.DefaultLogger;

import java.util.*;

import org.apache.log4j.Logger;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import  AITB_2023.module.util.NextCluster;


public class AITRoadDetector extends RoadDetector {
  /**
   * 自身の担当範囲のEntityのEntityIDの集合
   */
  private Set<EntityID> cluster = new HashSet<>();
  /**
   * 完了したタスクの集合
   */
  private Set<EntityID> completed = new HashSet<>();
  /**
   * PFの行動対象となる可能性のあるEntityのEntityIDの集合 key = PFの行動対象となる可能性のあるEntityのEntityID value =
   * keyのタスクの優先度（1（高）〜 8（低））
   */
  private Map<EntityID, Integer> tasks = new HashMap<>();

  private Set<Area> openedAreas = new HashSet<>();
  private Clustering clustering;
  private PathPlanning pathPlanning;

  private EntityID result;
  private Logger logger;

  /**
   * 通信用の定義
   */
  private final boolean VOICE = false;
  private final boolean RADIO = true;
  final private int HELP_BURIED = 5;
  final private int HELP_BLOCKADE = 6;

  // クラスタ関連
  private NextCluster nextCluster;
  private int myClusterIdx;
  private Collection<StandardEntity> myClusterEntities;
  private HashSet<Integer> searchedClusterIndexs;

  //救助要請を受けまだ探索していないリスト
  private List<EntityID> UnserchedList = new ArrayList<>();


  public AITRoadDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    this.pathPlanning = moduleManager.getModule(
        "SampleRoadDetector.PathPlanning",
        "adf.impl.module.algorithm.DijkstraPathPlanning");
    this.clustering = moduleManager.getModule("SampleRoadDetector.Clustering",
        "adf.impl.module.algorithm.KMeansClustring");
    registerModule(this.clustering);
    registerModule(this.pathPlanning);
    this.result = null;

    this.searchedClusterIndexs = new HashSet<>();
  }


  @Override
  public RoadDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    this.UnserchedList = new ArrayList<>();
    super.updateInfo(messageManager);
    this.receiveMessage(messageManager);
    this.sendMessage(messageManager);

    // 1stepに担当するクラスタの番号とクラスタ内のエンティティコレクションを取得
    if (this.agentInfo.getTime() == 1){
      // nextClusterを設定する
      this.nextCluster = new NextCluster(this.clustering.getAllClusterEntities(), this.agentInfo.me(),
              this.clustering.getClusterIndex(this.agentInfo.me()));
      //このエージェントが担当するクラスタの番号を取得し，myClusterIdxに代入
      this.myClusterIdx = this.clustering.getClusterIndex(this.agentInfo.getID());
      //idx番のクラスタに属するEntityをすべて取得し，myClusterEntitiesに代入
      this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
      // 探索クラスタリストに初期クラスタIDを追加
      this.searchedClusterIndexs.add(this.myClusterIdx);
    }
    return this;
  }


  @Override
  public RoadDetector calc() {
    //土木隊の現在位置する建物や道路のIDの取得
    EntityID positionID = this.agentInfo.getPosition();
    StandardEntity currentPosition = worldInfo.getEntity(positionID);
    //openedAreasには土木隊の現在位置のエンティティが格納される
    openedAreas.add((Area) currentPosition);
    //土木隊エージェントが前ステップから移動していないとき(おそらく瓦礫などに埋もれているとき)
    if (positionID.equals(result)) {
      logger.debug("reach to " + currentPosition + " resetting target");
      this.result = null;
    }

    //ターゲットが決まっていないときcalcTargets()へ
    if (this.result == null) {
      HashSet<Area> currentTargets = calcTargets();
      logger.debug("Targets: " + currentTargets);
      //ターゲットが指定されていないときresultをリセット
      if (currentTargets.isEmpty()) {
        this.result = null;
        return this;
      }
      this.pathPlanning.setFrom(positionID);
      this.pathPlanning.setDestination(toEntityIds(currentTargets));
      List<EntityID> path = this.pathPlanning.calc().getResult();
      if (path != null && path.size() > 0) {
        //pathのインデックスが大きい順番に格納していく
        this.result = path.get(path.size() - 1);
      }
      logger.debug("Selected Target: " + this.result);
    }
    return this;
  }


  private Collection<EntityID>
      toEntityIds(Collection<? extends StandardEntity> entities) {
    ArrayList<EntityID> eids = new ArrayList<>();
    for (StandardEntity standardEntity : entities) {
      eids.add(standardEntity.getID());
    }
    return eids;
  }

  //探索するエリアを決定する関数
  private HashSet<Area> calcTargets() {
    HashSet<Area> targetAreas = new HashSet<>();
    HashSet<Area> targetCivAreas = new HashSet<>();
    HashSet<Area> targetAgentAreas = new HashSet<>();
    HashSet<Area> targetComAreas = new HashSet<>();
    HashSet<Area> inClusterTarget = new HashSet<>();
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(REFUGE)) {
      targetCivAreas.add((Area) e);
    }
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(CIVILIAN)) {
      //建物に埋没している市民を探索対象に加える
      if (isValidCivilian(e)) {
        Human c = (Human) e;
        targetCivAreas.add((Area) worldInfo.getEntity(c.getPosition()));
      }
    }
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(AMBULANCE_TEAM, FIRE_BRIGADE)) {
      //瓦礫，建物に埋没しているエージェントを探索対象に加える
      if (isValidAgent(e)) {
        Human a = (Human) e;
        targetAgentAreas.add((Area) worldInfo.getEntity(a.getPosition()));
      }
    }

    //通信で受け取った位置情報を探索対象に加える
    for (EntityID unserched : this.UnserchedList) {
      targetComAreas.add((Area) worldInfo.getEntity(unserched));
    }

//    while (true) {
//      if (this.UnserchedList.isEmpty()) {
//        break;
//      }
//      targetAreas.add((Area) worldInfo.getEntity(this.UnserchedList.get(this.UnserchedList.size() - 1)));
//      this.UnserchedList.remove(this.UnserchedList.size()-1);
//    }

//    if (targetComAreas.isEmpty()) {
//      targetAreas = targetCivAreas;
//      HashSet<Area> inClusterTarget2 = filterInCluster(targetAreas);
//      inClusterTarget2.removeAll(openedAreas);
//      if (!(inClusterTarget2.isEmpty())) {
//        inClusterTarget = inClusterTarget2;
//      }
//    } else if (targetCivAreas.isEmpty()) {
//      targetAreas = targetAgentAreas;
//      HashSet<Area> inClusterTarget3 = filterInCluster(targetAreas);
//      inClusterTarget3.removeAll(openedAreas);
//      if (!(inClusterTarget3.isEmpty())) {
//        inClusterTarget = inClusterTarget3;
//      }
//    } else {
//      targetAreas = targetComAreas;
//      HashSet<Area> inClusterTarget1 = filterInCluster(targetAreas);
//      //目的地にたどり着いたとき，その場所を探索対象のリストから削除する
//      inClusterTarget1.removeAll(openedAreas);
//      if (!(inClusterTarget1.isEmpty())) {
//        inClusterTarget = inClusterTarget1;
//      }
//    }


    if (!(targetComAreas.isEmpty())) {
      targetAreas = targetComAreas;
      HashSet<Area> inClusterTarget1 = filterInCluster(targetAreas);
      //目的地にたどり着いたとき，その場所を探索対象のリストから削除する
      inClusterTarget1.removeAll(openedAreas);
      if (!(inClusterTarget1.isEmpty())) {
        inClusterTarget = inClusterTarget1;
      }
    } else if (!(targetCivAreas.isEmpty())) {
      targetAreas = targetCivAreas;
      HashSet<Area> inClusterTarget2 = filterInCluster(targetAreas);
      inClusterTarget2.removeAll(openedAreas);
      if (!(inClusterTarget2.isEmpty())) {
        inClusterTarget = inClusterTarget2;
      }
    } else {
      targetAreas = targetAgentAreas;
      HashSet<Area> inClusterTarget3 = filterInCluster(targetAreas);
      inClusterTarget3.removeAll(openedAreas);
      if (!(inClusterTarget3.isEmpty())) {
        inClusterTarget = inClusterTarget3;
      }
    }

    return inClusterTarget;
  }

  //自分のクラスタ内に埋まってる市民orエージェントがいたらそれらをclusterTargetsに格納
  private HashSet<Area> filterInCluster(HashSet<Area> targetAreas) {
    //あるエージェントが割当られたクラスタ番号
    int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    HashSet<Area> clusterTargets = new HashSet<>();
    //引数にとったクラスタ内の全てのエンティティ(道路や建物)
    HashSet<StandardEntity> inCluster = new HashSet<>(
            clustering.getClusterEntities(clusterIndex));
    for (Area target : targetAreas) {
      if (inCluster.contains(target))
        clusterTargets.add(target);
    }
    return clusterTargets;
  }


  @Override
  public EntityID getTarget() {
    return this.result;
  }


  //市民が建物に埋まっているときに(道路の瓦礫は含まれない)calcTargets()に探索対象の候補として追加してもらう関数
  private boolean isValidCivilian(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    //死んでるとき除く
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    //ダメージを受けてないとき除く(道路の瓦礫に埋もれているときは助けない)
    if (!target.isDamageDefined() || target.getDamage() == 0)
      return false;
    if (!target.isBuriednessDefined() || target.getBuriedness() == 0)
      return false;

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    //避難所にいる，または運ばれているとき除く
    StandardEntityURN positionURN = position.getStandardURN();
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM)
      return false;

    return true;
  }

  //エージェントが瓦礫や建物に埋まっているときにcalcTargetsに探索候補として追加してもらう関数
  private boolean isValidAgent(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    //死んでるとき除く
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    if (!target.isBuriednessDefined() || target.getBuriedness() == 0)
      return false;

    StandardEntity posi = this.worldInfo.getEntity(target.getPosition());
    if (!(posi instanceof Area)){
      return false;
    }

    PF_mes(" getBlockades: "+((Area) posi).getBlockades());
    //道路上に瓦礫がなかったら除く
    if (((Area) posi).getBlockades() == null) {
      return false;
    }

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    return true;
  }

  /**
   * 自身が埋没しているときメッセージを送信して助けを求める
   * @param mm MessageManager
   */
  private void sendMessage(MessageManager mm) {
    final PoliceForce me = (PoliceForce) this.worldInfo.getEntity(this.agentInfo.getID());
    // 自身が埋没しているとき
    if (this.isMyselfBuried()) {
      mm.addMessage(new MessagePoliceForce(RADIO, me, HELP_BURIED, this.agentInfo.getID()));
      mm.addMessage(new MessagePoliceForce(VOICE, me, HELP_BURIED, this.agentInfo.getID()));
      return;
    }
  }

  private boolean isMyselfBuried() {
    Human me = (Human) this.worldInfo.getEntity(this.agentInfo.getID());
    if (!(this.agentInfo.getPositionArea() instanceof Building)) return false;
    if (!me.isBuriednessDefined()) return false;
    return me.getBuriedness() != 0;
  }


  /**
   * エージェントが受け取ったメッセージを整理し，各々のメッセージをthis.handleMessage()で処理をする
   * @param mm MessageManager
   */
  private void receiveMessage(MessageManager mm) {
    //AmbulanceTeamが瓦礫で動けないというメッセージを取得
    final List<CommunicationMessage> helpAmbulance = mm.getReceivedMessageList(MessageAmbulanceTeam.class);
    String tt = "AT\n";
    for(CommunicationMessage mes : helpAmbulance) {
      final MessageAmbulanceTeam message = (MessageAmbulanceTeam) mes;

      EntityID sender = message.getSenderID();
      StandardEntity senderEntity = this.worldInfo.getEntity(sender);
      EntityID targetPosition = message.getTargetID();
      if (targetPosition != null) {
        this.UnserchedList.add(targetPosition);
      }
      int actionAT = message.getAction();
      tt+=" sender:"+sender+" targetPosition:"+targetPosition+"action:"+actionAT+"\n";

      this.handleMessage(message);
    }
    PF_mes(tt);
    tt = "F\n";
    //FireBrigadeが瓦礫で動けないというメッセージを取得
    final List<CommunicationMessage> helpFire = mm.getReceivedMessageList(MessageFireBrigade.class);
    for(CommunicationMessage mes : helpFire) {
      final MessageFireBrigade message = (MessageFireBrigade) mes;

      EntityID sender = message.getSenderID();
      StandardEntity senderEntity = this.worldInfo.getEntity(sender);
      EntityID targetPosition = message.getTargetID();
      if (targetPosition != null) {
        this.UnserchedList.add(targetPosition);
      }
      int actionFB = message.getAction();
      tt+=" sender:"+sender+" targetPosition:"+targetPosition+"action:"+actionFB+"\n";

      this.handleMessage(message);
    }
    PF_mes(tt);
  }

  /**
   * MessageFireBrigadeから自身のタスク候補へ追加
   * @param message エージェントが受け取ったMessageFireBrigade
   */
  private void handleMessage(MessageFireBrigade message) {
    final EntityID sender = message.getSenderID();
    final StandardEntity senderEntity = this.worldInfo.getEntity(sender);
    final EntityID targetPosition = message.getTargetID();

    if(message.getAction() != HELP_BLOCKADE) return;
    if(!message.isTargetDefined()) return;

    if(this.worldInfo.getDistance(this.agentInfo.getID(), sender)
            <= this.scenarioInfo.getRawConfig().getIntValue("comms.channels.0.range")) {
      this.putTask(targetPosition, 0);
      this.completed.remove(targetPosition);
    }

    if(!this.cluster.contains(targetPosition)) return;

    this.putTask(targetPosition, 0);
    this.completed.remove(targetPosition);
  }

  /**
   * MessageAmbulanceから自身のタスク候補へ追加
   * @param message エージェントが受け取ったMessageAmbulance
   */
  private void handleMessage(MessageAmbulanceTeam message) {
    final EntityID sender = message.getSenderID();
    final StandardEntity senderEntity = this.worldInfo.getEntity(sender);
    final EntityID targetPosition = message.getTargetID();

    if(message.getAction() != HELP_BLOCKADE) return;
    if(!message.isTargetDefined()) return;

    if(this.worldInfo.getDistance(this.agentInfo.getID(), sender)
            <= this.scenarioInfo.getRawConfig().getIntValue("comms.channels.0.range")) {
      this.putTask(targetPosition, 0);
      this.completed.remove(targetPosition);
    }

    if(!this.cluster.contains(targetPosition)) return;

    this.putTask(targetPosition, 0);
    this.completed.remove(targetPosition);
  }

  /**
   * PFの行動対象候補となるEntityIDを優先度とともに登録する
   *
   * @param task     PFの行動対象候補であるEntityのEntityID
   * @param priority taskの優先度（1（高）〜8（低））のInteger
   */
  private void putTask(EntityID task, int priority) {
    final int current = this.tasks.getOrDefault(task, Integer.MAX_VALUE);
    this.tasks.put(task, Math.min(current, priority));
  }

  // 担当クラスタを変更するメソッド(AITSearchでAITRoadDetectorのクラスタを変更する関数)
  public void change_cluster(int idx, HashSet<Integer> scIndexs){
    this.myClusterIdx = idx;
    this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
    this.searchedClusterIndexs = new HashSet<>(scIndexs);
    PF_mes(" クラスタを変更しました idx="+this.myClusterIdx+"\n"+this.searchedClusterIndexs);
  }


  /**
   *logをみるための関数
   *
   */
  public void PF_mes (String mes) {
    System.out.println("time: "+this.agentInfo.getTime()+" ID: "+this.agentInfo.getID().getValue()+mes);
  }

}
