package AITB_2023.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.log4j.Logger;

import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.StandardMessagePriority;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.communication.CommunicationMessage;
//import AITB_2023.module.complex.ATHumanDetector;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.complex.Search;
import adf.core.debug.DefaultLogger;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

public class AITSearch extends Search {
    private EntityID result;
    private Random random = new Random(1);
    private Clustering clustering;
    private List<EntityID> unsearchedBuildingIDs;
    private MessageManager messageManager;
    private static final int ACTION_LOAD = CommandAmbulance.ACTION_LOAD;
    private Logger logger;

    // 現在のクラスタID
    private int myClusterIdx;
    private Collection<StandardEntity> myClusterEntities;

    public AITSearch(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd) {
        super(ai, wi, si, mm, dd);
        this.logger = DefaultLogger.getLogger(agentInfo.me());

        String clusteringKey = "SampleSearch.Clustering";
        switch (ai.me().getStandardURN()) {
            case FIRE_BRIGADE:
                clusteringKey += ".Fire";
                break;
            case AMBULANCE_TEAM:
                clusteringKey += ".Ambulance";
                break;
            case POLICE_FORCE:
                clusteringKey += ".Police";
                break;
        }

        this.clustering = mm.getModule(clusteringKey,
                "adf.sample.module.algorithm.SampleKMeans");
        this.registerModule(this.clustering);

        this.unsearchedBuildingIDs = new ArrayList<>();
    }

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public Search updateInfo(MessageManager mm) {
        super.updateInfo(mm);
        // 別のモジュールでregisterModuleを書いたとき、２回目以降は実行されなくなる
        if (this.getCountUpdateInfo() > 1) return this;
        logger.debug("time: "+this.agentInfo.getTime()+" Search_updateInfo");

        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages = mm.getReceivedMessageList();
        this.messageManager = mm;

        // 辿り着いた対象を未探索IDコレクションから除外する
        if(getTarget()!=null) {
            EntityID position = this.agentInfo.getPosition();// エージェントが現在位置する建物や道路のIDを取得する
            if(this.getTarget().equals(position)) this.unsearchedBuildingIDs.remove(getTarget());
        }

        // もし未探索IDコレクションに要素がなかったら
        if (this.unsearchedBuildingIDs.isEmpty()) {
            this.reset();// リセットする
            if (getTarget() != null)
                this.unsearchedBuildingIDs.remove(getTarget());
        }

        return this;
    }

    @Override
    public Search calc()
    {
        logger.debug("time: "+this.agentInfo.getTime()+" Search_calc");
        if(this.myClusterIdx == -1) return this;// クラスタIDが-1なら終了

        if (this.myClusterEntities == null)
            return this;// クラスタEntitiesがnullなら終了

        // 視覚内の搬送されていない市民のメッセージ送信
        yetLoadCivilian_send();

        // エージェント自身のEntityIDを取得
        EntityID me = this.agentInfo.getID();
        // エージェントが現在位置する建物や道路のIDを取得する
        EntityID position = this.agentInfo.getPosition();

        // 最初の対象をランダムに選択する
        if (this.getTarget() == null) {

            // 擬似乱数を用いてランダムに選択します
            int n = this.unsearchedBuildingIDs.size();
            if (n > 0) {
                int r = this.random.nextInt(n);
                this.result = this.unsearchedBuildingIDs.get(r);
            }
            return this;
        }

        // 対象に辿り着くまで，対象を変更しない
        if (!this.getTarget().equals(position)) {
            return this;
        }

        double near = Double.POSITIVE_INFINITY;// nearesDistanceに初期値としてDoubleで最も大きい値を代入
        // 一番距離が近い建物を探索
        for (EntityID id : this.unsearchedBuildingIDs) {
            StandardEntity test = this.worldInfo.getEntity(id);
            if (test instanceof Building) {
                double distance = this.worldInfo.getDistance(me, id);
                if (distance < near) {
                    near = distance;
                    // System.out.println("ok");
                    this.result = id;
                }
            }
        }
        return this;
    }

    // 探索コレクションをリセットするメソッド
    private void reset() {
        this.unsearchedBuildingIDs.clear();// 未探索IDコレクションの要素を全て消す
        // 担当クラスタを決定
        if (this.agentInfo.getTime() == 1) {// もしstepが1なら
            // 自分が担当するクラスタのクラスタIDを取得
            this.myClusterIdx = this.clustering.getClusterIndex(this.agentInfo.getID());
            // クラスタIDからクラスタ内のエンティティを取得
            this.myClusterEntities = this.clustering.getClusterEntities(this.myClusterIdx);
            // System.out.println("クラスタ初期設定: "+this.myClusterIdx+"
            // "+this.myClusterEntities);
        } else {
            // logger.debug("Time: "+this.agentInfo.getTime()+" id:
            // "+this.agentInfo.getID()+"担当クラスタを変更判断");
            // change_cluster();// 担当クラスタを変更する処理
        }
        // クラスタ内エンティティの要素を未探索IDコレクションにaddする
        if (this.myClusterEntities != null && this.myClusterEntities.size() > 0) {// クラスタ内エンティティコレクションがnullでなく、要素数が0以上のとき
            for (StandardEntity entity : this.myClusterEntities) {// 拡張for
                // もしエンティティがBuildingクラスかつEntityのenumがREFUGEでないなら
                if (entity instanceof Building && entity.getStandardURN() != REFUGE) {
                    this.unsearchedBuildingIDs.add(entity.getID());// 未探索コレクションにEntityIDを追加
                }
            }
        } else {
            this.unsearchedBuildingIDs
                    .addAll(this.worldInfo.getEntityIDsOfType(BUILDING));
            /*
             * .addAll(this.worldInfo.getEntityIDsOfType(BUILDING, GAS_STATION,
             * AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE));
             */
        }
    }

    // 視覚内の搬送されていない市民のメッセージを送信するメソッド
    private void yetLoadCivilian_send() {
        // 視覚内にいる搬送されていない市民のIDリストを取得
        List<EntityID> yetLoadCivilians = this.worldInfo.getChanged().getChangedEntities()
                .stream()
                .map(this.worldInfo::getEntity)
                .filter(Civilian.class::isInstance)
                .map(Civilian.class::cast)
                .filter(Civilian::isDamageDefined)
                .filter(e -> e.getDamage() > 0) // ダメージあり
                .filter(Civilian::isHPDefined)
                .filter(e -> e.getHP() > 0) // 死亡していない
                .filter(Civilian::isBuriednessDefined)
                .filter(e -> e.getBuriedness() == 0) // 埋没度0
                .map(Civilian::getID)
                .collect(Collectors.toList());

        // メッセージの送信
        for (EntityID civilian : yetLoadCivilians) {
            Civilian civ = (Civilian) this.worldInfo.getEntity(civilian);
            messageManager.addMessage(new MessageCivilian(false, StandardMessagePriority.HIGH, civ), false);
        }
    }

    // 担当クラスタを変更するメソッド
    /*
     * private void change_cluster(){
     * ATHumanDetector atHumanDetector = (ATHumanDetector)this.humanDetector;
     * if(atHumanDetector.isBuriedCivilian()){
     * System.out.println("time: "+this.agentInfo.getTime()
     * +" id: "+this.agentInfo.getID()+" クラスタを変更しませんでした");
     * logger.debug("time: "+this.agentInfo.getTime()
     * +" id: "+this.agentInfo.getID()+" クラスタを変更しませんでした");
     * return; // もしクラスタ内に埋まっている市民がいるなら
     * }
     * int n = this.clustering.getClusterNumber();
     * this.myClusterIdx = (this.myClusterIdx + 1) % n;// クラスタidを1加算
     * this.myClusterEntities =
     * this.clustering.getClusterEntities(this.myClusterIdx);//
     * クラスタIDからクラスタ内のエンティティを取得
     * // 消防隊humanDetectorの方も変える
     * atHumanDetector.change_cluster(this.myClusterIdx);
     * //logger.debug("time: "+this.agentInfo.getTime() + " idx: "+
     * this.myClusterIdx + " " + this.myClusterEntities);
     * logger.debug("time: "+this.agentInfo.getTime()+" id: "+this.agentInfo.getID()
     * + " クラスタを変更しました ");
     * //System.out.println("time: "+this.agentInfo.getTime()+" id: "+this.agentInfo
     * .getID() + " idx: "+ this.myClusterIdx + " " + this.myClusterEntities);
     * System.out.println("time: "+this.agentInfo.getTime()+" id: "+this.agentInfo.
     * getID() + " クラスタを変更しました ");
     * }
     */
}