package AITB_2023.module.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.stream.Collectors;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;

public class NextCluster {
  private List<Collection<StandardEntity>> clusters;
  private Integer preClusInteger;
  private StandardEntity me;
  private String str;

  public NextCluster(List<Collection<StandardEntity>> clusters, StandardEntity me, Integer preClusInteger) {
    if (clusters == null || me == null || preClusInteger == null)
      throw new IllegalArgumentException("Argument is null");
    this.clusters = clusters;
    this.me = me;
    this.preClusInteger = preClusInteger;
  }

  public void setClusters(List<Collection<StandardEntity>> clusters) {
    this.clusters = clusters;
  }

  public void setMe(Human me) {
    this.me = me;
  }

  public void setPreClusInteger(Integer preClusInteger) {
    this.preClusInteger = preClusInteger;
  }

  public Integer getNextClusterIndex(HashSet<Integer> serachedClusterIndexs) {
    if (!serachedClusterIndexs.contains(preClusInteger))
      return preClusInteger;

    List<Integer> distance = clusters.stream()
        .map(this::getClusterDistance)
        .map(Double::intValue)
        .collect(Collectors.toList());

    Integer result = -1;
    Integer minDistance = Integer.MAX_VALUE;
    this.str = "";
    for (Integer i = 0; i < distance.size(); i++) {
      str += ("["+i + ": "+distance.get(i)+"]");
      if (distance.get(i) < minDistance && !serachedClusterIndexs.contains(i)) {
        minDistance = distance.get(i);
        result = i;
      }
    }
    str += ("\n result: "+result);

    return result;
  }

  public Integer getBuildingInNextClusterIndex(HashSet<Integer> serachedClusterIndexs) {
    if (!serachedClusterIndexs.contains(preClusInteger))
      return preClusInteger;

    List<Integer> distance = clusters.stream()
        .filter(this::isInBuilding)
        .map(this::getClusterDistance)
        .map(Double::intValue)
        .collect(Collectors.toList());

    Integer result = -1;
    Integer minDistance = Integer.MAX_VALUE;
    this.str = "";
    for (Integer i = 0; i < distance.size(); i++) {
      str += ("["+i + ": "+distance.get(i)+"]");
      if (distance.get(i) < minDistance && !serachedClusterIndexs.contains(i)) {
        minDistance = distance.get(i);
        result = i;
      }
    }
    str += ("\n result: "+result);

    return result;
  }

  private double getClusterDistance(Collection<StandardEntity> entities) {
    double x = 0, y = 0;
    for (StandardEntity entity : entities) {
      Area area = (Area) entity;
      x += area.getX();
      y += area.getY();
    }
    x /= entities.size();
    y /= entities.size();
    Human me = (Human) this.me;
    return Math.sqrt(Math.pow(x - me.getX(), 2) + Math.pow(y - me.getY(), 2));
  }

  private boolean isInBuilding(Collection<StandardEntity> entities){
    Collection<StandardEntity> inBuildingEntities =  entities.stream()
    .filter(entity -> entity instanceof Building)
    .filter(entity -> !(entity instanceof Refuge))
    .collect(Collectors.toSet());

    boolean inBuilding = (inBuildingEntities.isEmpty() ? false : true);

    return inBuilding;
  }

  public String getStr(){
    return this.str;
  }
}
